@@ -11,13 +11,13 @@ Cache::FastMmap - Uses an mmap'ed file to act as a shared memory interprocess ca
   # Uses vaguely sane defaults
   $Cache = Cache::FastMmap->new();
 
-  # $Value must be a reference...
+  # Uses Storable to serialize $Value to bytes for storage
   $Cache->set($Key, $Value);
   $Value = $Cache->get($Key);
 
-  $Cache = Cache::FastMmap->new(raw_values => 1);
+  $Cache = Cache::FastMmap->new(serializer => '');
 
-  # $Value can't be a reference...
+  # Stores stringified bytes of $Value directly
   $Cache->set($Key, $Value);
   $Value = $Cache->get($Key);
 
@@ -293,7 +293,7 @@ use strict;
 use warnings;
 use bytes;
 
-our $VERSION = '1.43';
+our $VERSION = '1.48';
 
 require XSLoader;
 XSLoader::load('Cache::FastMmap', $VERSION);
@@ -303,6 +303,9 @@ XSLoader::load('Cache::FastMmap', $VERSION);
 our %LiveCaches;
 
 use constant FC_ISDIRTY => 1;
+
+use File::Spec;
+
 # }}}
 
 =item I<new(%Opts)>
@@ -330,21 +333,84 @@ consistent file structure. The shared file is not perfectly transaction
 safe, and so if a child is killed at the wrong instant, it might leave
 the cache file in an inconsistent state.
 
+=item * B<serializer>
+
+Use a serialization library to serialize perl data structures before
+storing in the cache. If not set, the raw value in the variable passed
+to set() is stored as a string. You must set this if you want to store
+anything other than basic scalar values. Supported values are:
+
+  ''         for none
+  'storable' for 'Storable'
+  'sereal'   for 'Sereal'
+  'json'     for 'JSON'
+  [ $s, $d ] for custom serializer/de-serializer
+
+If this parameter has a value the module will attempt to load the
+associated package and then use the API of that package to serialize data
+before storing in the cache, and deserialize it upon retrieval from the
+cache. (default: 'storable')
+
+You can use a custom serializer/de-serializer by passing an array-ref
+with two values. The first should be a subroutine reference that takes
+the data to serialize as a single argument and returns an octet stream
+to store. The second should be a subroutine reference that takes the
+octet stream as a single argument and returns the original data structure.
+
+One thing to note, the data structure passed to the serializer is always
+a *scalar* reference to the original data passed in to the ->set(...)
+call. If your serializer doesn't support that, you might need to
+dereference it first before storing, but rembember to return a reference
+again in the de-serializer.
+
+(Note: Historically this module only supported a boolean value for the
+`raw_values` parameter and defaulted to 0, which meant it used Storable
+to serialze all values.)
+
 =item * B<raw_values>
 
-Store values as raw binary data rather than using Storable to free/thaw
-data structures (default: 0)
+Deprecated. Use B<serializer> above
+
+=item * B<compressor>
+
+Compress the value (but not the key) before storing into the cache, using
+the compression package identified by the value of the parameter. Supported
+values are:
+
+  'zlib'     for 'Compress::Zlib'
+  'lz4'      for 'Compress::LZ4'
+  'snappy'   for 'Compress::Snappy'
+  [ $c, $d ] for custom compressor/de-compressor
+
+If this parameter has a value the module will attempt to load the
+associated package and then use the API of that package to compress data
+before storing in the cache, and uncompress it upon retrieval from the
+cache. (default: undef)
+
+You can use a custom compressor/de-compressor by passing an array-ref
+with two values. The first should be a subroutine reference that takes
+the data to compress as a single octet stream argument and returns an
+octet stream to store. The second should be a subroutine reference that
+takes the compressed octet stream as a single argument and returns the
+original uncompressed data.
+
+(Note: Historically this module only supported a boolean value for the
+`compress` parameter and defaulted to use Compress::Zlib. The note for the
+old `compress` parameter stated: "Some initial testing shows that the
+uncompressing tends to be very fast, though the compressing can be quite
+slow, so it's probably best to use this option only if you know values in
+the cache are long-lived and have a high hit rate."
+
+Comparable test results for the other compression tools are not yet available;
+submission of benchmarks welcome. However, the documentation for the 'Snappy'
+library (http://google.github.io/snappy/) states: For instance, compared to
+the fastest mode of zlib, Snappy is an order of magnitude faster for most
+inputs, but the resulting compressed files are anywhere from 20% to 100%
+bigger. )
 
 =item * B<compress>
 
-Compress the value (but not the key) before storing into the cache. If
-you set this to 1, the module will attempt to require the Compress::Zlib
-module and then use the memGzip() function on the value data before
-storing into the cache, and memGunzip() when retrieving data from the
-cache. Some initial testing shows that the uncompressing tends to be
-very fast, though the compressing can be quite slow, so it's probably
-best to use this option only if you know values in the cache are long
-lived and have a high hit rate. (default: 0)
+Deprecated. Please use B<compressor>, see above.
 
 =item * B<enable_stats>
 
@@ -510,13 +576,13 @@ sub new {
   # Work out cache file and whether to init
   my $share_file = $Args{share_file};
   if (!$share_file) {
-    my $tmp_dir = $ENV{TMPDIR} || "/tmp";
-    my $win_tmp_dir = $ENV{TEMP} || "c:\\";
-    $share_file = ($^O eq "MSWin32" ? "$win_tmp_dir\\sharefile" : "$tmp_dir/sharefile");
+    my $tmp_dir = File::Spec->tmpdir;
+    $share_file = File::Spec->catfile($tmp_dir, "sharefile");
     $share_file .= "-" . $$ . "-" . time . "-" . int(rand(100000));
   }
   !ref($share_file) || die "share_file argument was a reference";
   $Self->{share_file} = $share_file;
+  my $permissions = $Args{permissions};
 
   my $init_file = $Args{init_file} ? 1 : 0;
   my $test_file = $Args{test_file} ? 1 : 0;
@@ -528,27 +594,73 @@ sub new {
     $Args{unlink_on_exit} = -f($share_file) ? 0 : 1;
   }
 
-  # Storing raw/storable values?
-  my $raw_values = $Self->{raw_values} = int($Args{raw_values} || 0);
+  # Serialise stored values?
+  my $serializer = $Args{serializer};
+  $serializer = ($Args{raw_values} ? '' : 'storable') if !defined $serializer;
 
-  # Need storable module if not using raw values
-  if (!$raw_values) {
-    eval "use Storable qw(freeze thaw); 1;"
-      || die "Could not load Storable module: $@";
+  if ($serializer) {
+    if (ref $serializer eq 'ARRAY') {
+      $Self->{serialize}   = $serializer->[0];
+      $Self->{deserialize} = $serializer->[1];
+    } elsif ($serializer eq 'storable') {
+      eval "require Storable;"
+        || die "Could not load serialization package: Storable : $@";
+      $Self->{serialize}   = Storable->can("freeze");
+      $Self->{deserialize} = Storable->can("thaw");
+    } elsif ($serializer eq 'sereal') {
+      eval "require Sereal::Encoder; require Sereal::Decoder;"
+        || die "Could not load serialization package: Sereal : $@";
+      my $SerealEnc = Sereal::Encoder->new();
+      my $SerealDec = Sereal::Decoder->new();
+      $Self->{serialize} = sub { $SerealEnc->encode(@_); };
+      $Self->{deserialize} = sub { $SerealDec->decode(@_); };
+    } elsif ($serializer eq 'json') {
+      eval "require JSON;"
+        || die "Could not load serialization package: JSON : $@";
+      my $JSON = JSON->new->utf8->allow_nonref;
+      $Self->{serialize}   = sub { $JSON->encode(${$_[0]}); };
+      $Self->{deserialize} = sub { \$JSON->decode($_[0]); };
+    } else {
+      die "Unrecognized value >$serializer< for `serializer` parameter";
+    }
   }
 
   # Compress stored values?
-  my $compress = $Self->{compress} = int($Args{compress} || 0);
+  my $compressor = $Args{compressor};
+  $compressor = ($Args{compress} ? 'zlib' : 0) if !defined $compressor;
 
-  # Need Compress::Zlib module if using compression
-  if ($compress) {
-    eval "use Compress::Zlib; 1;"
-      || die "Could not load Compress::Zlib module: $@";
+  my %known_compressors = (
+    zlib   => 'Compress::Zlib',
+    lz4    => 'Compress::LZ4',
+    snappy => 'Compress::Snappy',
+  );
+
+  if ( $compressor ) {
+    if (ref $compressor eq 'ARRAY') {
+      $Self->{compress}   = $compressor->[0];
+      $Self->{uncompress} = $compressor->[1];
+    } elsif (my $compressor_module = $known_compressors{$compressor}) {
+      eval "require $compressor_module;"
+        || die "Could not load compression package: $compressor_module : $@";
+
+      # LZ4 and Snappy use same API
+      if ($compressor_module eq 'Compress::LZ4' || $compressor_module eq 'Compress::Snappy') {
+        $Self->{compress}   = $compressor_module->can("compress");
+        $Self->{uncompress} = $compressor_module->can("uncompress");
+      } elsif ($compressor_module eq 'Compress::Zlib') {
+        $Self->{compress}   = $compressor_module->can("memGzip");
+        # (gunzip from tmp var: https://rt.cpan.org/Ticket/Display.html?id=72945)
+        my $uncompress = $compressor_module->can("memGunzip");
+        $Self->{uncompress} = sub { &$uncompress(my $Tmp = shift) };
+      }
+    } else {
+      die "Unrecognized value >$compressor< for `compressor` parameter";
+    }
   }
 
   # If using empty_on_exit, need to track used caches
   my $empty_on_exit = $Self->{empty_on_exit} = int($Args{empty_on_exit} || 0);
-  
+
   # Need Scalar::Util::weaken to track open caches
   if ($empty_on_exit) {
     eval "use Scalar::Util qw(weaken); 1;"
@@ -626,12 +738,12 @@ sub new {
 
   # Setup cache parameters
   fc_set_param($Cache, 'init_file', $init_file);
-  fc_set_param($Cache, 'init_file', $init_file);
   fc_set_param($Cache, 'test_file', $test_file);
   fc_set_param($Cache, 'page_size', $page_size);
   fc_set_param($Cache, 'num_pages', $num_pages);
   fc_set_param($Cache, 'expire_time', $expire_time);
   fc_set_param($Cache, 'share_file', $share_file);
+  fc_set_param($Cache, 'permissions', $permissions) if defined $permissions;
   fc_set_param($Cache, 'start_slots', $start_slots);
   fc_set_param($Cache, 'catch_deadlocks', $catch_deadlocks);
   fc_set_param($Cache, 'enable_stats', $enable_stats);
@@ -688,9 +800,8 @@ sub get {
       # Are we doing writeback's? If so, need to mark as dirty in cache
       my $write_back = $Self->{write_back};
 
-      # If not using raw values, use freeze() to turn data 
-      $Val = freeze(\$Val) if !$Self->{raw_values};
-      $Val = Compress::Zlib::memGzip($Val) if $Self->{compress};
+      $Val = $Self->{serialize}(\$Val) if $Self->{serialize};
+      $Val = $Self->{compress}($Val) if $Self->{compress};
 
       # Get key/value len (we've got 'use bytes'), and do expunge check to
       #  create space if needed
@@ -707,9 +818,8 @@ sub get {
   $Unlock = undef unless $SkipUnlock;
 
   # If not using raw values, use thaw() to turn data back into object
-  # (gunzip from tmp var: https://rt.cpan.org/Ticket/Display.html?id=72945)
-  $Val = Compress::Zlib::memGunzip(my $Tmp = $Val) if defined($Val) && $Self->{compress};
-  $Val = ${thaw($Val)} if defined($Val) && !$Self->{raw_values};
+  $Val = $Self->{uncompress}($Val) if defined($Val) && $Self->{compress};
+  $Val = ${$Self->{deserialize}($Val)} if defined($Val) && $Self->{deserialize};
 
   # If explicitly asked to skip unlocking, we return the reference to the unlocker
   return ($Val, $Unlock) if $SkipUnlock;
@@ -733,9 +843,8 @@ for more details.
 sub set {
   my ($Self, $Cache) = ($_[0], $_[0]->{Cache});
 
-  # If not using raw values, use freeze() to turn data 
-  my $Val = $Self->{raw_values} ? $_[2] : freeze(\$_[2]);
-  $Val = Compress::Zlib::memGzip($Val) if $Self->{compress};
+  my $Val = $Self->{serialize} ? $Self->{serialize}(\$_[2]) : $_[2];
+  $Val = $Self->{compress}($Val) if $Self->{compress};
 
   # Get opts, make compatible with Cache::Cache interface
   my $Opts = defined($_[3]) ? (ref($_[3]) ? $_[3] : { expire_time => $_[3] }) : undef;
@@ -836,7 +945,7 @@ sub get_and_set {
   my $DidStore = 0;
   if (@NewValue) {
     ($Value) = @NewValue;
-    my $DidStore = $Self->set($_[1], $Value, { skip_lock => \$Unlock });
+    $DidStore = $Self->set($_[1], $Value, { skip_lock => \$Unlock });
   }
 
   return wantarray ? ($Value, $DidStore) : $Value;
@@ -958,10 +1067,10 @@ sub get_keys {
   my ($Self, $Cache) = ($_[0], $_[0]->{Cache});
 
   my $Mode = $_[1] || 0;
-  my ($Compress, $RawValues) = @$Self{qw(compress raw_values)};
+  my ($Uncompress, $Deserialize) = @$Self{qw(uncompress deserialize)};
 
   return fc_get_keys($Cache, $Mode)
-    if $Mode <= 1 || ($Mode == 2 && $RawValues && !$Compress);
+    if $Mode <= 1 || ($Mode == 2 && !$Uncompress && !$Deserialize);
 
   # If we're getting values as well, and they're not raw, unfreeze them
   my @Details = fc_get_keys($Cache, 2);
@@ -969,11 +1078,8 @@ sub get_keys {
   for (@Details) {
     my $Val = $_->{value};
     if (defined $Val) {
-      $Val = Compress::Zlib::memGunzip($Val) if $Compress;
-      if (!$RawValues) {
-        $Val = eval { thaw($Val) };
-        $Val = $$Val if ref($Val);
-      }
+      $Val = $Uncompress->($Val) if $Uncompress;
+      $Val = ${$Deserialize->($Val)} if $Deserialize;
       $_->{value} = $Val;
     }
   }
@@ -1068,8 +1174,8 @@ sub multi_get {
     next unless $Found;
 
     # If not using raw values, use thaw() to turn data back into object
-    $Val = Compress::Zlib::memGunzip($Val) if defined($Val) && $Self->{compress};
-    $Val = ${thaw($Val)} if defined($Val) && !$Self->{raw_values};
+    $Val = $Self->{uncompress}($Val) if defined($Val) && $Self->{compress};
+    $Val = ${$Self->{deserialize}($Val)} if defined($Val) && $Self->{deserialize};
 
     # Save to return
     $KVs{$_} = $Val;
@@ -1101,9 +1207,8 @@ sub multi_set {
   my $KVs = $_[2];
   while (my ($Key, $Val) = each %$KVs) {
 
-    # If not using raw values, use freeze() to turn data 
-    $Val = freeze(\$Val) unless $Self->{raw_values};
-    $Val = Compress::Zlib::memGzip($Val) if $Self->{compress};
+    $Val = $Self->{serialize}(\$Val) if $Self->{serialize};
+    $Val = $Self->{compress}($Val) if $Self->{compress};
 
     # Get key/value len (we've got 'use bytes'), and do expunge check to
     #  create space if needed
@@ -1169,18 +1274,15 @@ sub _expunge_page {
 
   my @WBItems = fc_expunge($Cache, $Mode, $write_cb ? 1 : 0, $Len);
 
-  my ($Compress, $RawValues) = @$Self{qw(compress raw_values)};
+  my ($Uncompress, $Deserialize) = @$Self{qw(uncompress deserialize)};
 
   for (@WBItems) {
     next if !($_->{flags} & FC_ISDIRTY);
 
     my $Val = $_->{value};
     if (defined $Val) {
-      $Val = Compress::Zlib::memGunzip($Val) if $Compress;
-      if (!$RawValues) {
-        $Val = eval { thaw($Val) };
-        $Val = $$Val if ref($Val);
-      }
+      $Val = $Uncompress->($Val) if $Uncompress;
+      $Val = ${$Deserialize->($Val)} if $Deserialize;
     }
     eval { $write_cb->($Self->{context}, $_->{key}, $Val, $_->{expire_time}); };
   }
@@ -1388,7 +1490,7 @@ Rob Mueller L<mailto:cpan@robm.fastmail.fm>
 
 =head1 COPYRIGHT AND LICENSE
 
-Copyright (C) 2003-2015 by FastMail Pty Ltd
+Copyright (C) 2003-2017 by FastMail Pty Ltd
 
 This library is free software; you can redistribute it and/or modify
 it under the same terms as Perl itself.
